//////////////////////////////////////////////////////////////////////////
//
// SAD.cpp - Block matching algorithm (SAD)
//
// Author: Alexey Lukin, 2005
//
//////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "SAD.h"


//////////////////////////////////////////////////////////////////////////
/*! Calculates SAD for 8x8 blocks
 *  Side effect: overwrites registers, flags, SSE registers */
uint SAD8x8_SSE(const byte *pSrc, const byte *pDst, uint uWidth)
{
	uint uSum;
	__asm
	{
		mov		esi, pSrc		; ds:[esi] points to the source image block
		mov		edi, pDst		; ds:[edi] points to the destination image block
		mov		ebx, uWidth		; ebx = uWidth
		mov		edx, ebx		; edx = uWidth
		shl		edx, 1			; edx = uWidth * 2

		; Load source rows in mm registers
		movq	mm0, [esi]
		movq	mm1, [esi+ebx]
		add		esi, edx
		movq	mm2, [esi]
		movq	mm3, [esi+ebx]
		add		esi, edx
		movq	mm4, [esi]
		movq	mm5, [esi+ebx]
		add		esi, edx
		movq	mm6, [esi]
		movq	mm7, [esi+ebx]

		; Calculate SADs with destination rows
		psadbw	mm0, [edi]
		psadbw	mm1, [edi+ebx]
		add		edi, edx
		psadbw	mm2, [edi]
		psadbw	mm3, [edi+ebx]
		add		edi, edx
		psadbw	mm4, [edi]
		psadbw	mm5, [edi+ebx]
		add		edi, edx
		psadbw	mm6, [edi]
		psadbw	mm7, [edi+ebx]

		; Sum all SADs
		paddusw	mm0, mm1
		paddusw	mm0, mm2
		paddusw	mm0, mm3
		paddusw	mm0, mm4
		paddusw	mm0, mm5
		paddusw	mm0, mm6
		paddusw	mm0, mm7

		movd	uSum, mm0		; store sum
		emms					; empty MMX state
		mov		eax, uSum		; function result: eax
	}
} // end SAD8x8_SSE()
